use anyhow::{anyhow, bail, Context, Result};
use galactica_util::to_radians;
use nalgebra::{Point2, Point3};
use std::{
	collections::{HashMap, HashSet},
	sync::Arc,
};

use crate::{util::Polar, Content, ContentBuildContext, ContentIndex, Sprite};

pub(crate) mod syntax {
	use serde::Deserialize;
	use std::collections::HashMap;

	use crate::ContentIndex;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct System {
		pub name: String,
		pub object: HashMap<ContentIndex, Object>,
	}

	#[derive(Debug, Deserialize)]
	pub struct Object {
		pub sprite: ContentIndex,
		pub position: Position,

		pub size: f32,

		pub radius: Option<f32>,
		pub angle: Option<f32>,
		pub landable: Option<bool>,
		pub name: Option<String>,
		pub desc: Option<String>,
		pub image: Option<ContentIndex>,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Position {
		Polar(PolarCoords),
		Cartesian(CoordinatesThree),
	}

	#[derive(Debug, Deserialize)]
	pub struct PolarCoords {
		pub center: CoordinatesTwo,
		pub radius: f32,
		pub angle: f32,
		pub z: f32,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum CoordinatesTwo {
		Label(ContentIndex),
		Coords([f32; 2]),
	}

	impl ToString for CoordinatesTwo {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_string(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum CoordinatesThree {
		Label(ContentIndex),
		Coords([f32; 3]),
	}

	impl ToString for CoordinatesThree {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_string(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}
}

// Processed data structs.
// These are exported.

/// Represents a star system
#[derive(Debug, Clone)]
pub struct System {
	/// This object's name
	pub display_name: String,

	/// This object's index
	pub index: ContentIndex,

	/// Objects in this system
	pub objects: HashMap<ContentIndex, Arc<SystemObject>>,
}

/// Represents an orbiting body in a star system
/// (A star, planet, moon, satellite, etc)
/// These may be landable and may be decorative.
/// System objects to not interact with the physics engine.
#[derive(Debug, Clone)]
pub struct SystemObject {
	/// This object's name
	pub display_name: Option<String>,

	/// This object's index
	pub index: ContentIndex,

	/// This object's sprite
	pub sprite: Arc<Sprite>,

	/// This object's size.
	/// Measured as height in game units.
	/// This value is scaled for distance
	/// (i.e, the z-value of position)
	pub size: f32,

	/// This object's position, in game coordinates,
	/// relative to the system's center (0, 0).
	pub pos: Point3<f32>,

	/// This object's sprite's angle, in radians
	pub angle: f32,

	/// If true, ships may land on this object
	pub landable: bool,

	/// The description of this object (shown on landed ui)
	pub desc: Option<String>,

	/// This object's image (shown on landed ui)
	pub image: Option<Arc<Sprite>>,
}

/// Helper function for resolve_position, never called on its own.
fn resolve_coordinates(
	objects: &HashMap<ContentIndex, syntax::Object>,
	cor: &syntax::CoordinatesThree,
	mut cycle_detector: HashSet<ContentIndex>,
) -> Result<Point3<f32>> {
	match cor {
		syntax::CoordinatesThree::Coords(c) => Ok((*c).into()),
		syntax::CoordinatesThree::Label(l) => {
			if cycle_detector.contains(l) {
				bail!(
					"Found coordinate cycle: `{}`",
					cycle_detector.iter().fold(String::new(), |sum, a| {
						if sum.is_empty() {
							a.to_string()
						} else {
							format!("{sum} -> {a}")
						}
					})
				);
			}
			cycle_detector.insert(l.clone());

			let p = match objects.get(l) {
				Some(p) => p,
				None => bail!("Could not resolve coordinate label `{l}`"),
			};
			Ok(resolve_position(&objects, &p, cycle_detector)
				.with_context(|| format!("in object {:#?}", l))?)
		}
	}
}

/// Given an object, resolve its position as a Point3.
fn resolve_position(
	objects: &HashMap<ContentIndex, syntax::Object>,
	obj: &syntax::Object,
	cycle_detector: HashSet<ContentIndex>,
) -> Result<Point3<f32>> {
	match &obj.position {
		syntax::Position::Cartesian(c) => Ok(resolve_coordinates(objects, &c, cycle_detector)?),
		syntax::Position::Polar(p) => {
			let three = match &p.center {
				syntax::CoordinatesTwo::Label(s) => syntax::CoordinatesThree::Label(s.clone()),
				syntax::CoordinatesTwo::Coords(v) => {
					syntax::CoordinatesThree::Coords([v[0], v[1], f32::NAN])
				}
			};
			let r = resolve_coordinates(&objects, &three, cycle_detector)?;
			let plane = Polar {
				center: Point2::new(r.x, r.y),
				radius: p.radius,
				angle: to_radians(p.angle),
			}
			.to_cartesian();
			Ok(Point3::new(plane.x, plane.y, p.z))
		}
	}
}

impl crate::Build for System {
	type InputSyntaxType = HashMap<String, syntax::System>;

	fn build(
		system: Self::InputSyntaxType,
		_build_context: &mut ContentBuildContext,
		content: &mut Content,
	) -> Result<()> {
		for (system_name, system) in system {
			let mut objects = HashMap::new();

			for (index, obj) in &system.object {
				let mut cycle_detector = HashSet::new();
				cycle_detector.insert(index.clone());

				let sprite = match content.sprites.get(&obj.sprite) {
					None => bail!(
						"In system `{}`: sprite `{}` doesn't exist",
						system_name,
						obj.sprite
					),
					Some(t) => t.clone(),
				};

				let image = match &obj.image {
					Some(x) => match content.sprites.get(x) {
						None => bail!(
							"In system `{}`: sprite `{}` doesn't exist",
							system_name,
							obj.sprite
						),
						Some(t) => Some(t.clone()),
					},
					None => None,
				};

				if obj.landable.unwrap_or(false) {
					if obj.name.is_none() {
						return Err(anyhow!("if an object is landable, it must have a name"))
							.with_context(|| format!("in object labeled `{}`", index))
							.with_context(|| format!("in system `{}`", system_name));
					}

					if obj.desc.is_none() {
						return Err(anyhow!(
							"if an object is landable, it must have a description"
						))
						.with_context(|| format!("in object labeled `{}`", index))
						.with_context(|| format!("in system `{}`", system_name));
					}

					if obj.image.is_none() {
						return Err(anyhow!("if an object is landable, it must have an image"))
							.with_context(|| format!("in object labeled `{}`", index))
							.with_context(|| format!("in system `{}`", system_name));
					}
				}

				objects.insert(
					index.clone(),
					Arc::new(SystemObject {
						index: index.clone(),
						sprite,
						image,
						pos: resolve_position(&system.object, &obj, cycle_detector)
							.with_context(|| format!("in object {:#?}", index))?,
						size: obj.size,
						angle: to_radians(obj.angle.unwrap_or(0.0)),
						landable: obj.landable.unwrap_or(false),
						display_name: obj.name.as_ref().map(|x| x.clone()),
						// TODO: better linebreaks, handle double spaces
						// Tabs
						desc: obj
							.desc
							.as_ref()
							.map(|x| x.replace("\n", " ").replace("<br>", "\n")),
					}),
				);
			}

			content.systems.insert(
				ContentIndex::new(&system_name),
				Arc::new(Self {
					index: ContentIndex::new(&system_name),
					display_name: system.name,
					objects,
				}),
			);
		}

		return Ok(());
	}
}
